/*
 * @(#)IFFStructView.java  2.2.1  2009-03-27
 *
 * Copyright (c) 1999-2007 Werner Randelshofer, Immensee, Switzerland.
 * All rights reserved.
 *
 * You may not use, copy or modify this file, except in compliance with the
 * license agreement you entered into with Werner Randelshofer.
 * For details see accompanying license terms.
 */
package ch.randelshofer.view;

import org.monte.media.AbortException;
import org.monte.media.ParseException;
import org.monte.media.iff.*;
import org.monte.media.binary.*;
import ch.randelshofer.util.*;

import java.awt.*;
import java.io.*;
import java.text.NumberFormat;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;

import javax.swing.*;
import javax.swing.tree.*;

/**
 * Displays the structure of an IFF file and parses well known chunks.
 *
 * @author Werner Randelshofer
 * @version 2.2.1 2009-03-07 File IFFChunks.txt is now encoded with UTF-8.
 * <br>2.2 2005-06-19 indexInParent property added. <br>2.1.1 2004-12-26
 * IFFModel left file IFFChunks.txt open. <br>2.1 2004-04-27 Display child count
 * in tree. <br>2.0.1 2003-04-01 Revised. <br>2.0 2000-10-10 Parses now several
 * well known IFF chunks. <br>1.0 1999-10-19
 */
public class IFFStructView extends AbstractStructView {

    private InputStream inputStream;

    protected static class IFFChunkNode
            implements AbstractStructView.StructNode {

        private static NumberFormat numFormat = NumberFormat.getInstance();
        private IFFModel model;
        private int type;
        private int id;
        private long size;
        private long offset;
        private byte[] data;
        private int childCount;
        private int indexInParent;
        private File file;
        private boolean isDataChunk;

        public IFFChunkNode(IFFModel model, int type, int id, boolean isDataChunk,long size, long offset,   byte[] data) {
            this.model = model;
            this.type = type;
            this.id = id;
            this.size = size;
            this.offset = offset;
            this.data = data;
            this.isDataChunk=isDataChunk;
        }

        public IFFChunkNode(IFFModel model, int type, int id, boolean isDataChunk,long size, long offset,  File file) {
            this.model = model;
            this.type = type;
            this.id = id;
            this.size = size;
            this.offset = offset;
            this.file = file;
            this.isDataChunk=isDataChunk;
        }

        @Override
        public String getName() {
            return model.getName(this);
        }

        public String getType() {
            return IFFParser.idToString(type);
        }

        @Override
        public String getId() {
            return IFFParser.idToString(id);
        }

        @Override
        public long getSize() {
            return size;
        }

        @Override
        public long getOffset() {
            return offset;
        }

        public void setChildCount(int cc) {
            childCount = cc;
        }

        @Override
        public String getDescription() {
            return model.getDescription(this);
        }

        @Override
        public StructTableModel getStructTableModel() {
            return model.getStructTableModel(this);
        }

        @Override
        public byte[] getRawData(long offset, int len) {
            if (len <= 0) {
                return new byte[0];
            }
            if (data != null) {
                if (offset == 0 && len == data.length) {
                    return data;
                } else {
                    byte[] bytes = new byte[len];
                    System.arraycopy(data, 0, bytes, (int) offset, Math.min(data.length, len));
                    return bytes;
                }
            } else if (file != null) {
                try {
                    byte[] bytes = new byte[len];
                    RandomAccessFile raf = null;
                    try {
                        raf = new RandomAccessFile(file, "r");
                        raf.seek(this.offset + offset);
                        raf.read(bytes);
                    } catch (IOException e) {
                        e.printStackTrace();
                        file = null;
                    } finally {
                        if (raf != null) {
                            try {
                                raf.close();
                            } catch (IOException ex) {
                                ex.printStackTrace();
                                file = null;
                            }
                        }
                    }
                    return bytes;
                } catch (OutOfMemoryError e) {
                    e.printStackTrace();
                    return null;
                }
            }
            return null;
        }

        @Override
        public String toString() {
            if (!isDataChunk) {
                return IFFParser.idToString(type) + " " + IFFParser.idToString(id) + " " + numFormat.format(size) + " (" + childCount + ")";
            } else {
                return IFFParser.idToString(id) + " " + numFormat.format(size);
            }
        }

        @Override
        public int getIndexInParent() {
            return indexInParent;
        }

        public void setIndexInParent(int cc) {
            indexInParent = cc;
        }

        @Override
        public boolean hasRawData() {
            return size>0 && (data != null || file != null);
        }

        @Override
        public java.util.List<BinaryPanel.Highlight> getRawDataHighlights() {
            return null;
        }

        @Override
        public File getFile() {
            return null;
        }
    }

    protected static class IFFModel
            extends DefaultTreeModel
            implements IFFVisitor {

        private static StructParser structParser;
        private DefaultMutableTreeNode decodeNode;
        private File file;

        public IFFModel(InputStream in, File file)
                throws IOException {
            super(new DefaultMutableTreeNode());
            this.file = file;
            DefaultMutableTreeNode root = (DefaultMutableTreeNode) getRoot();

            if (structParser == null) {
                BufferedReader reader = null;
                try {
                    reader = new BufferedReader(
                            new InputStreamReader(
                            IFFStructView.class.getResourceAsStream("IFFChunks.txt"), "UTF-8"));
                    structParser = (reader == null) ? new StructParser() : new StructParser(reader);
                } catch (NullPointerException e) {
                    InternalError error = new InternalError(e.toString());
                    error.initCause(e);
                    throw error;
                } catch (ParseException e) {
                    InternalError error = new InternalError(e.toString());
                    error.initCause(e);
                    throw error;
                } finally {
                    if (reader != null) {
                        reader.close();
                    }
                }
            }

            try {
                IFFParser iff = new IFFParser();
                iff.setReadData(false);
                decodeNode = root;
                iff.parse(in, this);
            } catch (ParseException e) {
                throw new IOException(e.getMessage());
            } catch (AbortException e) {
                throw new IOException(e.getMessage());
            } finally {
                in.close();
            }
        }

        public void registerChunks(IFFParser iff) {
        }

        @Override
        public void enterGroup(IFFChunk group) {
            IFFChunkNode groupNode = new IFFChunkNode(
                    this,
                    group.getID(),
                    group.getType(),
                    false,
                    group.getSize(),
                     group.getScan(),
                    (byte[]) null);
            groupNode.setIndexInParent(decodeNode.getChildCount());
            DefaultMutableTreeNode newNode = new DefaultMutableTreeNode(groupNode);
            decodeNode.add(newNode);
            decodeNode = newNode;
        }

        @Override
        public void leaveGroup(IFFChunk group) {
            IFFChunkNode groupNode = (IFFChunkNode) decodeNode.getUserObject();
            groupNode.setChildCount(decodeNode.getChildCount());
            decodeNode = (DefaultMutableTreeNode) decodeNode.getParent();
        }

        @Override
        public void visitChunk(IFFChunk group, IFFChunk chunk) {
//System.out.println("IFFStructView.visitChunk group ID:"+group.getId()+" type:"+group.getType()+" chunk:"+chunk.getId());
            IFFChunkNode chunkNode;
            if (file == null) {
                chunkNode = new IFFChunkNode(
                        this,
                        group.getType(),
                        chunk.getID(),
                        true,
                         chunk.getSize(),
                         chunk.getScan(),
                        chunk.getData());
            } else {
                chunkNode = new IFFChunkNode(
                        this,
                        group.getType(),
                        chunk.getID(),
                        true,
                         chunk.getSize(),
                         chunk.getScan(),
                        file);
            }

            chunkNode.setIndexInParent(decodeNode.getChildCount());
            decodeNode.add(new DefaultMutableTreeNode(chunkNode));
        }

        private String getMagic(IFFChunkNode chunkNode) {
            String type = chunkNode.getType();
            String id = chunkNode.getId();
            String magic = type + id;
            if (structParser.isMagicDeclared(magic) == false) {
//System.out.println("IFFStructView.getMagic(...): no magic for:"+magic);                
                if (IFFParser.isGroupID(IFFParser.stringToID(type))) {
                    magic = type + "????";
                } else {
                    magic = "????" + id;
                }
            }
            return magic;
        }

        public String getName(IFFChunkNode chunkNode) {
            String name = structParser.getName(getMagic(chunkNode));
            return (name == null) ? chunkNode.getType() + " " + chunkNode.getId() : name;
        }

        public String getDescription(IFFChunkNode chunkNode) {
            return structParser.getDescription(getMagic(chunkNode));
        }

        public StructTableModel getStructTableModel(IFFChunkNode chunkNode) {
            try {
                byte[] data = chunkNode.getRawData(0, (int) chunkNode.getSize());
                return (data == null) ? null : structParser.readStruct(getMagic(chunkNode), data);
            } catch (IOException e) {
//System.err.println("IFFStructView.IFFModel.getStructTableModel(...):"+e);
//e.printStackTrace();
                return null;
            }
        }
    }

    /**
     * Initializes the Form
     */
    public IFFStructView() {
        //initComponents ();
    }

    @Override
    public void setCommandContext(String verb, DataHandler dh)
            throws IOException {
        inputStream = new BufferedInputStream(dh.getInputStream());
        setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        final File file;
        if (dh.getDataSource() instanceof FileDataSource) {
            file = ((FileDataSource) dh.getDataSource()).getFile();
        } else {
            file = null;
        }

        ch.randelshofer.gui.SwingWorker worker = new ch.randelshofer.gui.SwingWorker() {
            @Override
            public Object construct() {
                Object result;

                try {
                    result = new IFFModel(inputStream,file);
                } catch (Throwable e) {
                    result = e;
                }
                try {
                    inputStream.close();
                } catch (IOException e) {
                }

                return result;
            }

            @Override
            public void finished() {
                Object result = get();
                if (result instanceof IFFModel) {
                    getTree().setModel((IFFModel) result);
                } else {
                    removeAll();
                    add(new JLabel(result.toString()));
                    validate();
                }
                setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
                getTree().expandRow(0);
                getTree().setSelectionRow(0);
                getSplitPane().setDividerLocation(getTree().getPreferredSize().width + 20);
            }
        };
        worker.start();
    }

    @Override
    public void dispose() {
        if (inputStream != null) {
            try {
                inputStream.close();
            } catch (IOException e) {
            }
            inputStream = null;
        }
    }

    @Override
    public String toString() {
        return "IFF Struct View";
    }

    /**
     * This method is called from within the constructor to initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is always
     * regenerated by the FormEditor.
     */
    private void initComponents () {//GEN-BEGIN:initComponents
        setLayout (new java.awt.BorderLayout ());

    }//GEN-END:initComponents
// Variables declaration - do not modify//GEN-BEGIN:variables
// End of variables declaration//GEN-END:variables
}
