/*
 * @(#)RTCPPacket.java
 * Created: 2005-04-21
 * Version: 2-0-alpha
 * Copyright (c) 2005-2006, University of Manchester All rights reserved. 
 * Andrew G D Rowley
 * Christian Vincenot <sipcom@cyberspace7.net>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. Redistributions in binary
 * form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials 
 * provided with the distribution. Neither the name of the University of 
 * Manchester nor the names of its contributors may be used to endorse or 
 * promote products derived from this software without specific prior written
 * permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package net.sf.fmj.media.rtp;

import java.io.IOException;
import java.net.DatagramPacket;

/**
 * Represents an RTCP Packet
 * 
 * @author Andrew G D Rowley
 * @version 1-1-alpha
 */
public class RTCPPacket {

    /**
     * RTCP SR Packet
     */
    public static final int PT_SR = 200;

    /**
     * RTCP RR Packet
     */
    public static final int PT_RR = 201;

    /**
     * RTCP SDES Packet
     */
    public static final int PT_SDES = 202;

    /**
     * RTCP BYE Packet
     */
    public static final int PT_BYE = 203;

    /**
     * RTCP APP Packet
     */
    public static final int PT_APP = 204;

    // The header of the packet
    private RTCPHeader header = null;

    // The packet data
    private byte[] data = null;

    /**
     * Creates a new RTCPPacket
     * @param packet The datagram to read the packet from
     * @throws IOException I/O Exception
     */
    public RTCPPacket(DatagramPacket packet) throws IOException {
        this(packet.getData(), packet.getOffset(), packet.getLength());
    }

    /**
     * Creates a new RTCPPacket
     * @param data The data to read the packet from
     * @param offset The offset into the data
     * @param length The length of the data
     * @throws IOException I/O Exception
     */
    public RTCPPacket(byte data[], int offset, int length) throws IOException {
        header = new RTCPHeader(data, offset, length);
        this.data = new byte[length - RTCPHeader.SIZE];
        System.arraycopy(data, offset + RTCPHeader.SIZE, this.data, 0, length
                - RTCPHeader.SIZE);
    }

    /**
     * Creates a new RTCPPacket
     * 
     * @param header
     *            The packet header
     * @param data
     *            The packet data
     */
    public RTCPPacket(RTCPHeader header, byte[] data) {
        this.header = header;
        this.data = data;
    }

    /**
     * Returns the RTCP header of the packet.
     * @return The header of the packet
     */
    public RTCPHeader getRtcpHeader() {
        return header;
    }

    /**
     * Returns the data contained in this packet.
     * @return The data of the packet
     */
    public byte[] getRtcpData() {
        return data;
    }

    /**
     * Returns the length of the payload (the data contained in the packet, following
     * the header).
     * @return The length of the data (not the header length field)
     */
    public int getRtcpLength() {
        return data.length;
    }
}